from typing import Dict, List, Mapping, Optional

from dependency_injector.containers import Container

from . import base, service


class Module(base.Module):
    def __init__(self, resources: Optional[Mapping[str, base.ResourceFactory]] = None,
                 services: Optional[Mapping[str, base.ServiceFactory]] = None,
                 modules: Optional[Mapping[str, base.ModuleFactory]] = None, *args, **kwargs):
        super(Module, self).__init__(*args, **kwargs)
        self._resources: Dict[str, base.Resource] = dict()
        self._services: Dict[str, service.Service] = dict()
        self._modules: Dict[str, Module] = dict()

        if resources:
            for name, provider in resources.items():
                instance = provider()
                if isinstance(instance, Container) and hasattr(instance, 'instance'):
                    instance = instance.instance()
                if isinstance(instance, base.Resource):
                    self._resources[name] = instance
        if services:
            for name, provider in services.items():
                instance = provider()
                if isinstance(instance, Container) and hasattr(instance, 'instance'):
                    instance = instance.instance()
                if isinstance(instance, base.Service):
                    self._services[name] = instance
        if modules:
            for name, provider in modules.items():
                instance = provider()
                if isinstance(instance, Container) and hasattr(instance, 'instance'):
                    instance = instance.instance()
                if isinstance(instance, base.Module):
                    self._modules[name] = instance

    @property
    def resources(self) -> List[base.Resource]:
        return list(self._resources.values())

    def get_resource(self, name) -> Optional[base.Resource]:
        return self._resources.get(name)

    @property
    def services(self) -> List[base.Service]:
        return list(self._services.values())

    def get_service(self, name) -> Optional[base.Service]:
        return self._services.get(name)

    @property
    def modules(self) -> List[base.Module]:
        return list(self._modules.values())

    def get_module(self, name) -> Optional[base.Module]:
        return self._modules.get(name)

    def sync_startup(self, application=None, *args, **kwargs):
        super().sync_startup(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            r.sync_startup(application=application, module=self, *args, **kwargs)
        for m in self.modules:
            m.sync_startup(application=application, module=self, *args, **kwargs)
        for s in self.services:
            s.sync_startup(application=application, module=self, *args, **kwargs)

    def sync_shutdown(self, application=None, *args, **kwargs):
        for s in self.services:
            s.sync_shutdown(application=application, module=self, *args, **kwargs)
        for m in self.modules:
            m.sync_shutdown(application=application, module=self, *args, **kwargs)
        for r in self.resources:
            r.sync_shutdown(application=application, module=self, *args, **kwargs)

        super().sync_shutdown(application=application, module=self, *args, **kwargs)

    async def startup(self, application=None, *args, **kwargs):
        await super().startup(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            await r.startup(application=application, module=self, *args, **kwargs)
        for m in self.modules:
            await m.startup(application=application, module=self, *args, **kwargs)
        for s in self.services:
            await s.startup(application=application, module=self, *args, **kwargs)

    async def shutdown(self, application=None, *args, **kwargs):
        for s in self.services:
            await s.shutdown(application=application, module=self, *args, **kwargs)
        for m in self.modules:
            await m.shutdown(application=application, module=self, *args, **kwargs)
        for r in self.resources:
            await r.shutdown(application=application, module=self, *args, **kwargs)

        await super().shutdown(application=application, module=self, *args, **kwargs)


class SessionModule(Module, base.SessionResource):
    def sync_session_begin(self, application=None, *args, **kwargs):
        super().sync_session_begin(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_begin(application=application, module=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResource):
                continue
            m.sync_session_begin(application=application, module=self, *args, **kwargs)

        for r in self.services:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_begin(application=application, module=self, *args, **kwargs)

    def sync_session_finish(self, application=None, *args, **kwargs):
        for r in self.services:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_finish(application=application, module=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResource):
                continue
            m.sync_session_finish(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_finish(application=application, module=self, *args, **kwargs)

        super().sync_session_finish(application=application, module=self, *args, **kwargs)

    async def session_begin(self, application=None, *args, **kwargs):
        await super().session_begin(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_begin(application=application, module=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResource):
                continue
            await m.session_begin(application=application, module=self, *args, **kwargs)

        for r in self.services:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_begin(application=application, module=self, *args, **kwargs)

    async def session_finish(self, application=None, *args, **kwargs):
        for r in self.services:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_finish(application=application, module=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResource):
                continue
            await m.session_finish(application=application, module=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_finish(application=application, module=self, *args, **kwargs)

        await super().session_finish(application=application, module=self, *args, **kwargs)



import contextlib
from pathlib import Path
from typing import Mapping, Optional

from dependency_injector import providers
from dependency_injector.containers import Container
from dependency_injector.wiring import Provide, inject

from dropland.data.context import with_context
from dropland.log import logger
from dropland.util import import_path, invoke_async, invoke_sync
from . import base, module


class Application(base.Application, module.SessionModule, base.ContainerResource):
    def __init__(self, container: Container, name: str, debug: bool = False, version: str = '',
                 resources: Optional[Mapping[str, base.ResourceFactory]] = None,
                 services: Optional[Mapping[str, base.ServiceFactory]] = None,
                 modules: Optional[Mapping[str, base.ModuleFactory]] = None):
        super().__init__(container=container, resources=resources, services=services, modules=modules)

        self._name = name
        self._version = version
        self._debug = debug
        self._path = self._inspect_cwd()

        self._imported_packages = None

        container.instance = providers.Object(self)
        container.wire(modules=[__name__])

    @property
    def name(self) -> str:
        return self._name

    @staticmethod
    def _inspect_cwd():
        import inspect
        path = Path(inspect.stack()[3].filename).absolute()
        result = None
        while path.parent != path and result is None:
            path = path.parent
            pyproject_toml_path = path / 'pyproject.toml'
            if pyproject_toml_path.exists():
                result = path
        return result

    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    def get_cwd(self) -> Path:
        return self._path.absolute()

    def sync_startup(self, *args, **kwargs):
        super().sync_startup(application=self, *args, **kwargs)

    def sync_shutdown(self, *args, **kwargs):
        super().sync_shutdown(application=self, *args, **kwargs)

    async def startup(self, *args, **kwargs):
        await super().startup(application=self, *args, **kwargs)

    async def shutdown(self, *args, **kwargs):
        await super().shutdown(application=self, *args, **kwargs)

    def sync_session_begin(self, *args, **kwargs):
        super().sync_session_begin(application=self, *args, **kwargs)

    def sync_session_finish(self, *args, **kwargs):
        super().sync_session_finish(application=self, *args, **kwargs)

    async def session_begin(self, *args, **kwargs):
        await super().session_begin(application=self, *args, **kwargs)

    async def session_finish(self, *args, **kwargs):
        await super().session_finish(application=self, *args, **kwargs)

    def _import_packages(self, key: Optional[str] = None):
        from tomlkit import parse

        packages, result = dict(), dict()
        pyproject_toml_path = self.get_cwd() / 'pyproject.toml'

        if pyproject_toml_path.exists():
            with open(str(pyproject_toml_path)) as f:
                pyproject_toml = parse(string=f.read())

            if project := pyproject_toml.get('project'):
                if key:
                    if section := project.get(key):
                        packages = section.get('packages', dict())
                else:
                    packages = project.get('packages', dict())

        key = key or 'package'

        for name, dotted_path in packages.items():
            logger.info(f'Import package: "{name}"')
            dotted_path = dotted_path if ':' in dotted_path else f'{dotted_path}:{key}_init'

            if entrypoint := import_path(dotted_path):
                result[name] = entrypoint

        return result

    def sync_load_packages(self, key, *args, **kwargs):
        if not self._imported_packages:
            self._imported_packages = self._import_packages(key)

        with self.sync_with_app_sessions():
            for name, entrypoint in self._imported_packages.items():
                logger.info(f'Application "{self.name}": load package "{name}"')
                invoke_sync(entrypoint, self, self.container, *args, **kwargs)

    async def load_packages(self, key, *args, **kwargs):
        if not self._imported_packages:
            self._imported_packages = self._import_packages(key)

        async with self.with_app_sessions():
            for name, entrypoint in self._imported_packages.items():
                logger.info(f'Application "{self.name}": load package "{name}"')
                await invoke_async(entrypoint, self, self.container, *args, **kwargs)

    @contextlib.contextmanager
    def sync_with_app_resources(self):
        with with_context(True):
            with contextlib.ExitStack() as stack:
                invoke_sync(self.startup)
                self.sync_startup()
                stack.callback(self.sync_shutdown)
                stack.callback(invoke_sync, self.shutdown)
                yield

    @contextlib.asynccontextmanager
    async def with_app_resources(self):
        with with_context(True):
            async with contextlib.AsyncExitStack() as stack:
                await self.startup()
                await invoke_async(self.sync_startup)
                stack.push_async_callback(invoke_async, self.sync_shutdown)
                stack.push_async_callback(self.shutdown)
                yield

    @contextlib.contextmanager
    def sync_with_app_sessions(self):
        with with_context(False):
            with contextlib.ExitStack() as stack:
                invoke_sync(self.session_begin)
                self.sync_session_begin()
                stack.callback(self.sync_session_finish)
                stack.callback(invoke_sync, self.session_finish)
                yield

    @contextlib.asynccontextmanager
    async def with_app_sessions(self):
        with with_context(False):
            async with contextlib.AsyncExitStack() as stack:
                await self.session_begin()
                await invoke_async(self.sync_session_begin)
                stack.push_async_callback(invoke_async, self.sync_session_finish)
                stack.push_async_callback(self.session_finish)
                yield


@inject
def current_application(_app: Application = Provide['instance']):
    return _app



try:
    import sqlalchemy.orm

    from .engine import EngineConfig, SqlEngineBackend, SqlEngine
    from .model import SqlaModelBase as SqlaModel

    USE_SQLA = True

except ImportError:
    USE_SQLA = False



from collections import OrderedDict
from datetime import timedelta
from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple

from dropland.data.cache import ModelCacheData, ModelCacheProtocol
from dropland.data.serializers import Deserializer, Serializer
from .engine import Connection, RedisEngine


class SimpleRedisModelCache(ModelCacheProtocol):
    def __init__(self, engine: RedisEngine, class_key: str,
                 serializer: Serializer, deserializer: Deserializer, ttl_enabled: bool = True):
        self._redis_engine = engine
        self._class_key = class_key
        self._serializer = serializer
        self._deserializer = deserializer
        self._ttl_enabled = ttl_enabled
        self._null_data = self._serializer.serialize(None)

    def get_model_cache_key(self) -> str:
        return f'{self._redis_engine.name}.models.{self._class_key}'

    async def cache_one(
            self, session: Connection, data: ModelCacheData, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        cache_kwargs = dict()

        if self._ttl_enabled:
            expiration = ttl or self._redis_engine.default_ttl
            total_seconds = expiration.total_seconds()

            if total_seconds > 1:
                cache_kwargs['expire'] = int(total_seconds)
            elif 0 < total_seconds < 1:
                cache_kwargs['pexpire'] = int(total_seconds * 1000)
            else:
                total_seconds = self._redis_engine.default_ttl.total_seconds()
                cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

        serialized = self._serializer.serialize(data.data)
        return bool(await session.set(self.get_cache_key(data.cache_id), serialized, **cache_kwargs))

    async def cache_many(
            self, session: Connection, objects: List[ModelCacheData],
            ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if not objects:
            return False
        res = False

        tx = session.multi_exec()

        for instance in objects:
            cache_kwargs = dict()

            if self._ttl_enabled:
                expiration = ttl or self._redis_engine.default_ttl
                total_seconds = expiration.total_seconds()

                if total_seconds > 1:
                    cache_kwargs['expire'] = int(total_seconds)
                elif 0 < total_seconds < 1:
                    cache_kwargs['pexpire'] = int(total_seconds * 1000)
                else:
                    total_seconds = self._redis_engine.default_ttl.total_seconds()
                    cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

            serialized = self._serializer.serialize(instance.data)
            tx.set(self.get_cache_key(instance.cache_id), serialized, **cache_kwargs)

        for r in await tx.execute():
            res |= r

        return bool(res)

    async def load_one(
            self, session: Connection, cache_key: str, **kwargs) -> Tuple[bool, Optional[Any]]:
        if res := await session.get(cache_key):
            return True, self._deserializer.deserialize(res)
        return False, None

    async def load_many(
            self, session: Connection, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []

        cache_keys = [self.get_cache_key(id_value) for id_value in indices]
        res = await session.mget(*cache_keys)
        objects: Dict[Any, Any] = OrderedDict()

        for id_value, data in zip(indices, res):
            objects[id_value] = self._deserializer.deserialize(data) if data is not None else None

        return list(objects.values())

    async def drop_one(self, session: Connection, cache_key: str) -> bool:
        return bool(await session.delete(cache_key))

    async def drop_many(self, session: Connection, indices: List[Any] = None) -> int:
        if indices and len(indices) > 0:
            cache_keys = [self.get_cache_key(id_value) for id_value in indices]
        else:
            model_cache_key = self.get_model_cache_key()
            cache_keys = await session.keys(f'{model_cache_key}:*')

        return int(await session.delete(*cache_keys)) if cache_keys else 0

    async def drop_all(self, session: Connection, prefix: Optional[str] = None) -> int:
        model_cache_key = self.get_model_cache_key()
        cache_key = f'{model_cache_key}:{prefix}*' if prefix else f'{model_cache_key}:*'
        keys_to_delete = await session.keys(cache_key)
        return int(await session.delete(*keys_to_delete)) if keys_to_delete else 0

    async def exists(self, session: Connection, cache_key: str) -> bool:
        if await session.exists(cache_key):
            return await session.get(cache_key) != self._null_data
        return False

    async def scan(self, session: Connection, cache_key: str = None,
                   match: str = None, count: int = None) \
            -> AsyncGenerator[Tuple[str, Optional[Dict[str, Any]]], None]:
        match = f'{cache_key}:{match}' if cache_key and match else match
        async for k in session.iscan(match=match, count=count):
            cache_key = k.decode('utf-8')
            k = cache_key.split(':')[1] if ':' in cache_key else cache_key
            v = (await self.load_one(session, cache_key))[1]
            yield k, v


class HashRedisModelCache(ModelCacheProtocol):
    def __init__(self, engine: RedisEngine, class_key: str,
                 serializer: Serializer, deserializer: Deserializer, ttl_enabled: bool = True):
        self._redis_engine = engine
        self._class_key = class_key
        self._serializer = serializer
        self._deserializer = deserializer
        self._ttl_enabled = ttl_enabled
        self._null_data = self._serializer.serialize(None)

    def get_model_cache_key(self) -> str:
        return f'{self._redis_engine.name}.models.{self._class_key}'

    async def cache_one(
            self, session: Connection, data: ModelCacheData, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        model_cache_key, cache_id = self.get_model_cache_key(), self.get_cache_id(data.cache_id)
        serialized = self._serializer.serialize(data.data)

        if self._ttl_enabled:
            res = False
            tx = session.multi_exec()
            expiration = ttl or self._redis_engine.default_ttl
            tx.hset(model_cache_key, cache_id, serialized)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await session.hset(model_cache_key, cache_id, serialized))

    async def cache_many(
            self, session: Connection, objects: List[ModelCacheData],
            ttl: Optional[timedelta] = None, **kwargs) -> bool:
        if not objects:
            return False

        model_cache_key = self.get_model_cache_key()
        obj_dict = {
            self.get_cache_id(instance.cache_id): self._serializer.serialize(instance.data)
            for instance in objects
        }

        if self._ttl_enabled:
            res = False
            tx = session.multi_exec()
            expiration = ttl or self._redis_engine.default_ttl
            tx.hmset_dict(model_cache_key, obj_dict)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await session.hmset_dict(model_cache_key, obj_dict))

    async def load_one(
            self, session: Connection, cache_key: str, **kwargs) -> Tuple[bool, Optional[Any]]:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            if res := await session.hget(model_cache_key, cache_id):
                return True, self._deserializer.deserialize(res)
            return False, None
        else:
            if res := await session.get(cache_key):
                return True, self._deserializer.deserialize(res)
            return False, None

    async def load_many(
            self, session: Connection, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []

        cache_keys = [self.get_cache_id(id_value) for id_value in indices]
        res = await session.hmget(self.get_model_cache_key(), *cache_keys)
        objects: Dict[Any, Any] = OrderedDict()

        for id_value, data in zip(indices, res):
            objects[id_value] = self._deserializer.deserialize(data) if data is not None else None

        return list(objects.values())

    async def drop_one(self, session: Connection, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            return bool(await session.hdel(model_cache_key, cache_id))
        else:
            return bool(await session.delete(cache_key))

    async def drop_many(self, session: Connection, indices: List[Any] = None) -> int:
        model_cache_key = self.get_model_cache_key()

        if indices and len(indices) > 0:
            cache_keys = [self.get_cache_id(id_value) for id_value in indices]
        else:
            cache_keys = await session.hkeys(model_cache_key)

        return int(await session.hdel(model_cache_key, *cache_keys)) if cache_keys else 0

    async def drop_all(self, session: Connection, prefix: Optional[str] = None) -> int:
        model_cache_key = self.get_model_cache_key()
        if not prefix:
            keys_to_delete = await session.hkeys(model_cache_key)
        else:
            keys_to_delete = list()
            for cache_key in await session.hkeys(model_cache_key, encoding='ascii'):
                if cache_key.startswith(prefix):
                    keys_to_delete.append(cache_key)

        return int(await session.hdel(model_cache_key, *keys_to_delete)) if keys_to_delete else 0

    async def exists(self, session: Connection, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':', maxsplit=1)
            e = bool(await session.hexists(model_cache_key, cache_id))
        else:
            e = bool(await session.exists(cache_key))
        if e:
            return await session.get(cache_key) != self._null_data
        return False

    async def scan(self, session: Connection, cache_key: str = None,
                   match: str = None, count: int = None) \
            -> AsyncGenerator[Tuple[str, Optional[Dict[str, Any]]], None]:
        async for k, data in session.ihscan(cache_key, match=match, count=count):
            yield k.decode('utf-8'), self._deserializer.deserialize(data)



from dependency_injector import containers, providers

from app.module import Module
from .databases import USE_DB
from .redis import USE_REDIS
from .rmq import USE_RMQ
from .scheduler import USE_SCHEDULER
from .sqla import USE_SQLA

__engines__ = dict()


if USE_DB:
    from .databases.containers import DbResource
    __engines__['db'] = DbResource

if USE_REDIS:
    from .redis.containers import RedisResource
    __engines__['redis'] = RedisResource

if USE_RMQ:
    from .rmq.containers import RmqResource
    __engines__['rmq'] = RmqResource

if USE_SCHEDULER:
    from .scheduler.containers import SchedulerResource
    __engines__['scheduler'] = SchedulerResource

if USE_SQLA:
    from .sqla.containers import SqlaResource
    __engines__['sql'] = SqlaResource


class EnginesContainer(containers.DeclarativeContainer):
    __self__ = providers.Self()

    engines = providers.Aggregate(**{
        name: providers.Factory(class_) for name, class_ in __engines__.items()
    })


class EnginesModule(Module):
    pass




from collections import defaultdict
from typing import Any, Optional, Protocol, Sequence

from .base import ModelProtocol
from ..context import get_context


class ModelLocalCacheMixin(Protocol):
    @classmethod
    def _add_to_local_cache(cls, instances: Sequence[ModelProtocol]):
        ctx = get_context()

        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        for instance in instances:
            if instance is not None:
                ctx.cache[cls][instance.get_id_value()] = instance

    @classmethod
    def _get_from_local_cache(cls, indices: Sequence[Any]) -> Sequence[Optional[ModelProtocol]]:
        ctx = get_context()

        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        return [ctx.cache[cls].get(id_value) for id_value in indices]

    @classmethod
    def _drop_from_local_cache(cls, indices: Sequence[Any] = None) -> Sequence[bool]:
        ctx = get_context()

        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        if indices is None:
            ctx.cache[cls] = defaultdict(dict)
            return [True]

        return [ctx.cache[cls].pop(id_value, None) is None for id_value in indices]

    @classmethod
    def _has_in_local_cache(cls, indices: Sequence[Any]) -> Sequence[bool]:
        ctx = get_context()

        if not hasattr(ctx, 'cache'):
            ctx.cache = defaultdict(lambda: defaultdict(dict))

        return [id_value in ctx.cache[cls] for id_value in indices]

    # noinspection PyProtectedMember
    @classmethod
    async def _register_instances(cls, objects: Sequence[ModelProtocol]):
        await super()._register_instances(objects)
        cls._add_to_local_cache(objects)

    # noinspection PyProtectedMember
    @classmethod
    async def _unregister_indices(cls, indices: Sequence[Any] = None):
        cls._drop_from_local_cache(indices)
        await super()._unregister_indices(indices)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, List, Optional, Protocol, runtime_checkable

from dependency_injector.containers import Container


@runtime_checkable
class BaseResource(Protocol):
    pass


class Resource(BaseResource):
    def __init__(self, *args, **kwargs):
        self._initialized: bool = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    def sync_startup(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    def sync_shutdown(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    async def startup(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    async def shutdown(self, application=None, module=None, service=None, *args, **kwargs):
        pass


class SessionResource(Resource):
    def sync_session_begin(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    def sync_session_finish(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    async def session_begin(self, application=None, module=None, service=None, *args, **kwargs):
        pass

    async def session_finish(self, application=None, module=None, service=None, *args, **kwargs):
        pass


ResourceFactory = Callable[[], Resource]


class ContainerResource(Resource):
    def __init__(self, container: Container, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._container = container

    @property
    def container(self) -> Container:
        return self._container


class Service(Resource, ABC):
    name: str
    resources: List[Resource]

    @abstractmethod
    def get_resource(self, name) -> Optional[Resource]:
        ...


ServiceFactory = Callable[[], Service]


class Module(Resource, ABC):
    resources: List[Resource]
    services: List[Service]
    modules: List['Module']

    @abstractmethod
    def get_resource(self, name) -> Optional[Resource]:
        ...

    @abstractmethod
    def get_service(self, name) -> Optional[Service]:
        ...

    @abstractmethod
    def get_module(self, name) -> Optional['Module']:
        ...


ModuleFactory = Callable[[], Module]


class Application(Module, ABC):
    name: str

    @property
    @abstractmethod
    def version(self) -> str:
        ...

    @property
    @abstractmethod
    def debug(self) -> bool:
        ...

    @abstractmethod
    def get_cwd(self) -> Path:
        ...

    @abstractmethod
    def sync_load_packages(self, *args, **kwargs):
        ...

    @abstractmethod
    async def load_packages(self, *args, **kwargs):
        ...



from typing import Dict, List, Mapping, Optional

from dependency_injector.containers import Container

from . import base


class Service(base.Service):
    def __init__(self, name: str, resources: Optional[Mapping[str, base.ResourceFactory]] = None):
        super(Service, self).__init__()
        self._name = name
        self._resources: Dict[str, base.Resource] = dict()

        if resources:
            for name, provider in resources.items():
                instance = provider()
                if isinstance(instance, Container) and hasattr(instance, 'instance'):
                    instance = instance.instance()
                if isinstance(instance, base.Resource):
                    self._resources[name] = instance

    @property
    def name(self) -> str:
        return self._name

    @property
    def resources(self) -> List[base.Resource]:
        return list(self._resources.values())

    def get_resource(self, name) -> Optional[base.Resource]:
        return self._resources.get(name)

    def sync_startup(self, application=None, module=None, *args, **kwargs):
        super().sync_startup(application=application, module=module, service=self, *args, **kwargs)

        for r in self.resources:
            r.sync_startup(application=application, module=module, service=self, *args, **kwargs)

    def sync_shutdown(self, application=None, module=None, *args, **kwargs):
        for r in self.resources:
            r.sync_shutdown(application=application, module=module, service=self, *args, **kwargs)

        super().sync_shutdown(application=application, module=module, service=self, *args, **kwargs)

    async def startup(self, application=None, module=None, *args, **kwargs):
        await super().startup(application=application, module=module, service=self, *args, **kwargs)

        for r in self.resources:
            await r.startup(application=application, module=module, service=self, *args, **kwargs)

    async def shutdown(self, application=None, module=None, *args, **kwargs):
        for r in self.resources:
            await r.shutdown(application=application, module=module, service=self, *args, **kwargs)

        await super().shutdown(application=application, module=module, service=self, *args, **kwargs)


class SessionService(Service, base.SessionResource):
    def sync_session_begin(self, application=None, module=None, *args, **kwargs):
        super().sync_session_begin(application=application, module=module, service=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_begin(application=application, module=module, service=self, *args, **kwargs)

    def sync_session_finish(self, application=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            r.sync_session_finish(application=application, module=module, service=self, *args, **kwargs)

        super().sync_session_finish(application=application, module=module, service=self, *args, **kwargs)

    async def session_begin(self, application=None, module=None, *args, **kwargs):
        await super().session_begin(application=application, module=module, service=self, *args, **kwargs)

        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_begin(application=application, module=module, service=self, *args, **kwargs)

    async def session_finish(self, application=None, module=None, *args, **kwargs):
        for r in self.resources:
            if not isinstance(r, base.SessionResource):
                continue
            await r.session_finish(application=application, module=module, service=self, *args, **kwargs)

        await super().session_finish(application=application, module=module, service=self, *args, **kwargs)



from typing import Protocol, runtime_checkable


@runtime_checkable
class EntityBase(Protocol):
    pass


class Entity(EntityBase):
    class Meta:
        pass

    def __init__(self, **kwargs):
        super().__init__()
        self.__dict__.update(kwargs)


class ModelEntity(Entity):
    async def save(self) -> bool:
        return True



import asyncio
import sys
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AbstractAsyncContextManager, AbstractContextManager, asynccontextmanager, contextmanager
from dataclasses import dataclass, replace
from datetime import timedelta
from typing import Callable, Dict, List, Mapping, Optional, Set

from orjson import orjson
from sqlalchemy import create_engine as create_sync_engine
from sqlalchemy.ext.asyncio import AsyncSession, async_scoped_session, create_async_engine
from sqlalchemy.orm import Session, scoped_session, sessionmaker
from sqlalchemy.schema import MetaData

from dropland.data.context import get_context
from dropland.log import logger, tr
from ..base import AsyncEngine, EngineBackend, SyncEngine
from ..sql import NAMING_CONVENTION, SqlEngineType


@dataclass
class EngineConfig:
    url: str
    echo: bool = False
    pool_min_size: int = 1
    pool_max_size: int = 8
    pool_expire_seconds: int = 60
    pool_timeout_seconds: int = 15


@dataclass
class EngineKey:
    db_type: SqlEngineType
    is_async: bool

    def __eq__(self, other: 'EngineKey'):
        return (self.db_type, self.is_async) == (other.db_type, other.is_async)

    def __hash__(self):
        return hash((self.db_type, self.is_async))


@dataclass
class EngineWithDbType:
    db_type: SqlEngineType
    engine: 'SqlEngine'


class SqlEngineBackend(EngineBackend):
    def __init__(self):
        self._engines: Dict[str, 'SqlEngine'] = dict()
        self._engines_by_key: Dict[EngineKey, Set[str]] = defaultdict(set)
        self._engines_by_type: Dict[SqlEngineType, Set[str]] = defaultdict(set)

    @property
    def name(self) -> str:
        return 'sqla'

    # noinspection PyMethodOverriding
    def create_engine(self, name: str, config: EngineConfig, db_type: SqlEngineType, use_async: bool) -> 'SqlEngine':
        if engine := self._engines.get(name):
            return engine

        uri = config.url

        if use_async:
            if db_type == SqlEngineType.SQLITE:
                uri = uri.replace('sqlite', 'sqlite+aiosqlite', 1)
            elif db_type == SqlEngineType.POSTGRES:
                uri = uri.replace('postgresql', 'postgresql+asyncpg', 1)
            elif db_type == SqlEngineType.MYSQL:
                uri = uri.replace('mysql', 'mysql+aiomysql', 1)
        else:
            if db_type == SqlEngineType.MYSQL:
                uri = uri.replace('mysql', 'mysql+pymysql', 1)

        config = replace(config, url=uri)

        params = dict(
            echo=config.echo,
            echo_pool=config.echo,
            pool_pre_ping=True,
            pool_recycle=config.pool_expire_seconds,
            json_serializer=orjson.dumps,
            json_deserializer=orjson.loads,
        )

        if db_type in (SqlEngineType.POSTGRES, SqlEngineType.MYSQL):
            params.update(dict(
                pool_size=config.pool_min_size,
                max_overflow=config.pool_max_size - config.pool_min_size,
                pool_timeout=config.pool_timeout_seconds
            ))

        if use_async:
            engine = create_async_engine(config.url, **params)
            session_factory = sessionmaker(engine, autoflush=False, expire_on_commit=False, class_=AsyncSession)
        else:
            engine = create_sync_engine(config.url, **params)
            session_factory = sessionmaker(engine, autoflush=False)

        logger.info(tr('dropland.engines.sql.engine.created')
                    .format(name=self.name, db_type=db_type, async_=f'async={use_async}'))
        metadata = MetaData(bind=engine, naming_convention=NAMING_CONVENTION)
        engine_class = SqlAlchemyAsyncEngine if use_async else SqlAlchemySyncEngine

        engine = engine_class(
            self, name, engine, metadata, db_type, session_factory,
            timedelta(seconds=config.pool_timeout_seconds)
        )

        self._engines[name] = engine
        self._engines_by_key[EngineKey(db_type=db_type, is_async=use_async)].add(name)
        self._engines_by_type[db_type].add(name)
        return engine

    def get_engine(self, name: str) -> Optional['SqlEngine']:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def get_engines_for_type(self, db_type: SqlEngineType, is_async: bool) -> List['SqlEngine']:
        engine_names = self._engines_by_key[EngineKey(db_type=db_type, is_async=is_async)]
        return [self.get_engine(name) for name in engine_names]

    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, 'SqlEngine']:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines


class SqlEngine(ABC):
    def __init__(self, name: str, raw_engine, metadata, db_type: SqlEngineType, timeout: timedelta):
        self._name = name
        self._db_type = db_type
        self._timeout = timeout
        self._metadata = metadata
        self._engine = raw_engine.execution_options(timeout=int(timeout.total_seconds()))

    @property
    def name(self) -> str:
        return self._name

    @property
    def raw_engine(self):
        return self._engine

    @property
    def db_type(self):
        return self._db_type

    @property
    def timeout(self) -> timedelta:
        return self._timeout

    @property
    def metadata(self):
        return self._metadata

    @property
    @abstractmethod
    def connection_class(self):
        ...


class SqlAlchemySyncEngine(SyncEngine[Session], SqlEngine):
    def __init__(self, backend: SqlEngineBackend, name: str, engine, metadata,
                 db_type: SqlEngineType, session_factory, timeout: timedelta):
        SyncEngine.__init__(self, backend)
        SqlEngine.__init__(self, name, engine, metadata, db_type, timeout)

        self._session_factory = scoped_session(
            session_factory, scopefunc=lambda: get_context().session_id
        )
        self._lock = threading.Lock()
        self._counter = 0

    @property
    def connection_class(self):
        return Session

    @contextmanager
    def session(self, begin_tx: bool = True, force_rollback: bool = False) \
            -> Callable[..., AbstractContextManager[Session]]:
        if self._session_factory.registry.has():
            yield self._session_factory()
            return

        session: Session = self._session_factory()

        if begin_tx:
            with session.begin() as tx:
                yield session

                if sys.exc_info()[0] or force_rollback:
                    tx.rollback()
                else:
                    tx.commit()
        else:
            yield session

    def teardown_session(self):
        self._session_factory.remove()

    def start(self):
        with self._lock:
            if 0 == self._counter:
                logger.info(tr('dropland.engines.sql.engine.started')
                            .format(name=self.name, db_type=self.db_type, async_=f'async=False'))
            self._counter += 1

    def stop(self):
        with self._lock:
            if 1 == self._counter:
                self.raw_engine.dispose()
                logger.info(tr('dropland.engines.sql.engine.stopped')
                            .format(name=self.name, db_type=self.db_type, async_=f'async=False'))
            self._counter = max(self._counter - 1, 0)


class SqlAlchemyAsyncEngine(AsyncEngine[AsyncSession], SqlEngine):
    def __init__(self, backend: SqlEngineBackend, name: str, engine, metadata,
                 db_type: SqlEngineType, session_factory, timeout: timedelta):
        AsyncEngine.__init__(self, backend)
        SqlEngine.__init__(self, name, engine, metadata, db_type, timeout)

        self._session_factory = async_scoped_session(
            session_factory, scopefunc=lambda: get_context().session_id
        )
        self._lock = asyncio.Lock()
        self._counter = 0

    @property
    def connection_class(self):
        return AsyncSession

    @asynccontextmanager
    async def session(self, begin_tx: bool = True, force_rollback: bool = False) \
            -> Callable[..., AbstractAsyncContextManager[AsyncSession]]:
        if self._session_factory.registry.has():
            yield self._session_factory()
            return

        session: AsyncSession = self._session_factory()

        if begin_tx:
            async with session.begin() as tx:
                yield session

                if sys.exc_info()[0] or force_rollback:
                    await tx.rollback()
                else:
                    await tx.commit()
        else:
            yield session

    async def teardown_session(self):
        await self._session_factory.remove()

    async def start(self):
        async with self._lock:
            if 0 == self._counter:
                logger.info(tr('dropland.engines.sql.engine.started')
                            .format(name=self.name, db_type=self.db_type, async_=f'async=True'))
            self._counter += 1

    async def stop(self):
        async with self._lock:
            if 1 == self._counter:
                await self.raw_engine.dispose()
                logger.info(tr('dropland.engines.sql.engine.stopped')
                            .format(name=self.name, db_type=self.db_type, async_=f'async=True'))
            self._counter = max(self._counter - 1, 0)



from dependency_injector import containers, providers

from dropland.app.base import ContainerResource, SessionResource
from dropland.util import default_value
from .engine import EngineConfig, RmqEngineBackend


class RmqResource(ContainerResource, SessionResource):
    async def startup(self, *args, **kwargs):
        if self.initialized:
            return

        for engine in self.container.engine_factory().get_engines().values():
            if engine.is_async:
                await engine.start()
            else:
                engine.start()

        self._initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self.initialized:
            return

        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if engine.is_async:
                await engine.stop()
            else:
                engine.stop()

        self._initialized = False

    async def session_begin(self, *args, **kwargs):
        pass

    async def session_finish(self, *args, **kwargs):
        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if engine.is_async:
                await engine.teardown_session()
            else:
                engine.teardown_session()

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SingleRmqContainer) or 'SingleRmqContainer' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)


class RmqContainer(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(RmqEngineBackend)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    create_engine = providers.Factory(_create_engine, __self__)
    instance = providers.Singleton(RmqResource, __self__)


class SingleRmqContainer(RmqContainer):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                virtualhost=self.config.engine_config.virtualhost(),
                timeout_seconds=self.config.engine_config.
                    timeout_seconds.as_(default_value(int))(default=5),
                pool_max_connections=self.config.engine_config.
                    pool_max_connections.as_(default_value(int))(default=4),
                pool_max_channels_per_connection=self.config.engine_config.
                    pool_max_channels_per_connection.as_(default_value(int))(default=100)
            )
        return RmqContainer._create_engine(self, self.config.name(), engine_config)

    create_engine = providers.Factory(_create_engine, __self__)
    instance = providers.Singleton(RmqResource, __self__)


class MultipleRmqContainer(RmqContainer):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    virtualhost=conf['engine_config']['virtualhost'],
                    timeout_seconds=int(conf['engine_config'].get('timeout_seconds', 5)),
                    pool_max_connections=int(conf['engine_config'].get('pool_max_connections', 4)),
                    pool_max_channels_per_connection=int(conf['engine_config'].get(
                        'pool_max_channels_per_connection', 100))
                )
            return RmqContainer._create_engine(self, name, engine_config)
        return None

    create_engine = providers.Factory(_create_engine, __self__)
    instance = providers.Singleton(RmqResource, __self__)



import asyncio
from collections import defaultdict
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from dataclasses import dataclass, replace
from typing import Dict, Set, Optional, List, Mapping, Callable, Any

from databases.core import Database, Connection

from dropland.engines.base import EngineBackend, AsyncEngine
from dropland.engines.sql import SqlEngineType
from dropland.log import tr, logger


@dataclass
class EngineConfig:
    url: str
    echo: bool = False
    pool_min_size: int = 1
    pool_max_size: int = 8
    pool_expire_seconds: int = 60


class DbEngineBackend(EngineBackend):
    def __init__(self):
        self._engines: Dict[str, 'DbEngine'] = dict()
        self._engines_by_type: Dict[SqlEngineType, Set[str]] = defaultdict(set)

    @property
    def name(self) -> str:
        return 'db'

    # noinspection PyMethodOverriding
    def create_engine(self, name: str, config: EngineConfig, db_type: SqlEngineType) -> 'DbEngine':
        if engine := self._engines.get(name):
            return engine

        uri = config.url
        params = dict()

        if db_type == SqlEngineType.SQLITE:
            uri = uri.replace('sqlite', 'sqlite+aiosqlite', 1)
        elif db_type == SqlEngineType.POSTGRES:
            uri = uri.replace('postgresql', 'postgresql+asyncpg', 1)
            params.update(dict(
                min_size=config.pool_min_size,
                max_size=config.pool_max_size,
                max_inactive_connection_lifetime=config.pool_expire_seconds,
            ))
        elif db_type == SqlEngineType.MYSQL:
            uri = uri.replace('mysql', 'mysql+aiomysql', 1)
            params.update(dict(
                echo=config.echo,
                minsize=config.pool_min_size,
                maxsize=config.pool_max_size,
                pool_recycle=config.pool_expire_seconds,
            ))

        config = replace(config, url=uri)
        engine = DbEngine(self, name, db_type, config, params)
        self._engines[name] = engine
        self._engines_by_type[db_type].add(name)
        logger.info(tr('dropland.engines.db.engine.created').format(name=name, db_type=db_type))
        return engine

    def get_engine(self, name: str) -> Optional['DbEngine']:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def get_engines_for_type(self, db_type: SqlEngineType) -> List['DbEngine']:
        engine_names = self._engines_by_type[db_type]
        return [self.get_engine(name) for name in engine_names]

    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, 'DbEngine']:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines


class DbEngine(AsyncEngine[Connection]):
    def __init__(self, backend: DbEngineBackend, name: str, db_type: SqlEngineType,
                 config: EngineConfig, params: Dict[str, Any]):
        super().__init__(backend)
        self._name = name
        self._db_type = db_type
        self._db = Database(config.url, **params)
        self._lock = asyncio.Lock()
        self._counter = 0

    @property
    def name(self) -> str:
        return self._name

    @property
    def db_type(self):
        return self._db_type

    @asynccontextmanager
    async def session(self, begin_tx: bool = True, force_rollback: bool = False) \
            -> Callable[..., AbstractAsyncContextManager[Connection]]:
        async with self._db.connection() as conn:
            if begin_tx:
                async with conn.transaction(force_rollback=force_rollback):
                    yield conn
            else:
                yield conn

    async def start(self):
        async with self._lock:
            if 0 == self._counter:
                await self._db.connect()
                logger.info(tr('dropland.engines.db.engine.started').format(name=self.name, db_type=self.db_type))
            self._counter += 1

    async def stop(self):
        async with self._lock:
            if 1 == self._counter:
                await self._db.disconnect()
                logger.info(tr('dropland.engines.db.engine.stopped').format(name=self.name, db_type=self.db_type))
            self._counter = max(self._counter - 1, 0)



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Dict, Generic, Optional, Sequence, Tuple, Type, TypeVar, Union

from .entities import Entity
from .models.base import ModelProtocol
from .models.model import UpdateSchemaType
from .models.sql import SqlModel

EntityType = TypeVar('EntityType', bound=Entity, covariant=True)
ModelType = TypeVar('ModelType', bound=ModelProtocol, covariant=True)
SqlModelType = TypeVar('SqlModelType', bound=SqlModel, covariant=True)
ModelTypeFactory = Callable[[...], ModelType]
SqlModelTypeFactory = Callable[[...], SqlModelType]


class Repository(Generic[EntityType], ABC):
    @abstractmethod
    def __call__(self, *args, **kwargs) -> Union[EntityType, Awaitable[EntityType]]:
        ...

    @abstractmethod
    async def get(self, id_value: Any, **kwargs) -> Optional[EntityType]:
        ...

    @abstractmethod
    async def get_any(self, indices: Sequence[Any], **kwargs) -> Sequence[Optional[EntityType]]:
        ...

    @abstractmethod
    async def exists(self, id_value: Any, **kwargs) -> bool:
        ...

    @abstractmethod
    async def create(self, data: Any, **kwargs) -> Optional[EntityType]:
        ...

    @abstractmethod
    async def get_or_create(self, id_value: Any, data: Any, **kwargs) -> Tuple[Optional[EntityType], bool]:
        ...

    @abstractmethod
    async def update_by_id(self, id_value: Any, data: Any) -> bool:
        ...

    @abstractmethod
    async def save_all(self, objects: Sequence[EntityType], *args, **kwargs) -> bool:
        ...

    @abstractmethod
    async def delete_all(self, indices: Sequence[Any] = None) -> bool:
        ...


class ModelRepository(Repository[ModelType]):
    def __init__(self, model_class: Type[ModelType], model_factory: Optional[ModelTypeFactory] = None):
        self._model_class = model_class
        self._model_factory = model_factory or model_class

    def __call__(self, *args, **kwargs) -> Union[ModelType, Awaitable[ModelType]]:
        return self._model_factory(*args, **kwargs)

    @property
    def model_class(self):
        return self._model_class

    #
    # Retrieve operations
    #

    async def get(self, id_value: Any, **kwargs) -> Optional[ModelType]:
        return await self._model_class.get(id_value, **kwargs)

    async def get_any(self, indices: Sequence[Any], **kwargs) -> Sequence[Optional[ModelType]]:
        return await self._model_class.get_any(indices, **kwargs)

    async def exists(self, id_value: Any, **kwargs) -> bool:
        return await self._model_class.exists(id_value, **kwargs)

    #
    # Modification operations
    #

    async def create(self, data: Any, **kwargs) -> Optional[ModelType]:
        return await self._model_class.create(data, **kwargs)

    async def get_or_create(self, id_value: Any, data: Any, **kwargs) -> Tuple[Optional[ModelType], bool]:
        return await self._model_class.get_or_create(id_value, data, **kwargs)

    async def update_by_id(self, id_value: Any, data: Any) -> bool:
        return await self._model_class.update_by_id(id_value, data)

    async def save_all(self, objects: Sequence[ModelType], *args, **kwargs) -> bool:
        return await self._model_class.save_all(objects, *args, **kwargs)

    async def delete_all(self, indices: Sequence[Any] = None) -> bool:
        return await self._model_class.delete_all(indices)


class SqlModelRepository(ModelRepository[SqlModelType]):
    def __init__(self, model_class: Type[SqlModelType], model_factory: Optional[SqlModelTypeFactory] = None):
        super(SqlModelRepository, self).__init__(model_class, model_factory)

    async def list(
        self, filters: Optional[Sequence[Any]] = None, sorting: Optional[Sequence[Any]] = None,
            skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs) -> Sequence[SqlModelType]:
        return await self._model_class.list(filters, sorting, skip, limit, params, **kwargs)

    async def count(self, filters: Optional[Sequence[Any]] = None, params: Dict[str, Any] = None, **kwargs) -> int:
        return await self._model_class.count(filters, params, **kwargs)

    async def exists_by(self, filters: Sequence[Any], params: Dict[str, Any] = None, **kwargs) -> bool:
        return await self._model_class.exists_by(filters, params, **kwargs)

    async def update_by(self, filters: Sequence[Any],
                        data: Union[UpdateSchemaType, Dict[str, Any]],
                        /, params: Dict[str, Any] = None, **kwargs) -> int:
        return await self._model_class.update_by(filters, data, params, **kwargs)

    async def delete_by(self, filters: Sequence[Any], /, params: Dict[str, Any] = None, **kwargs) -> int:
        return await self._model_class.delete_by(filters, params, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dropland',
 'dropland.app',
 'dropland.core',
 'dropland.core.loaders',
 'dropland.data',
 'dropland.data.models',
 'dropland.data.serializers',
 'dropland.engines',
 'dropland.engines.databases',
 'dropland.engines.redis',
 'dropland.engines.rmq',
 'dropland.engines.scheduler',
 'dropland.engines.sqla',
 'dropland.ext',
 'dropland.ext.fastapi',
 'dropland.tr']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils>=0.37.2,<0.38.0',
 'contextvars>=2.4,<3.0',
 'dependency-injector[pydantic]>=4.40.0',
 'orjson>=3.8.1,<4.0.0',
 'poetry-version>=0.2.0,<0.3.0',
 'pydantic[email,dotenv]>=1.8.2,<2.0.0',
 'pytz>=2021.3,<2022.0',
 'timeparse-plus>=1.2.0,<2.0.0',
 'tomlkit<0.6.0']

extras_require = \
{':extra == "fastapi"': ['fastapi[all]>=0.85.0'],
 'db': ['alembic>=1.8.0,<2.0.0', 'databases>=0.6.1'],
 'mysql': ['aiomysql>=0.0.21', 'pymysql[rsa]>=0.9,<=0.9.3'],
 'pg': ['asyncpg>=0.22.0,<0.23.0', 'psycopg2>=2.8.6,<3.0.0'],
 'redis': ['aioredis[hiredis]<=1.3.1', 'redis>=4.1.0'],
 'rmq': ['aio-pika>=8.2.0'],
 'scheduler': ['APScheduler>=3.7.0', 'rpyc>=5.1.0'],
 'sqla': ['SQLAlchemy[asyncio,mypy]>=1.4,<2.0', 'alembic>=1.8.0,<2.0.0'],
 'sqlite': ['aiosqlite>=0.17.0,<0.18.0']}

setup_kwargs = {
    'name': 'dropland',
    'version': '0.8.2',
    'description': 'Mini-framework for building a backend servers for web-services using SQLAlchemy, Databases, Redis, RabbitMQ and APScheduler',
    'long_description': 'Dropland\n========\n\nMini-framework for building a backend servers for web-services using SQLAlchemy, Databases, Redis, RabbitMQ and APScheduler\n\n\nHow to build\n------------\n\n- Create a Python virtual environment.\n\n    ``pyenv local 3.9.0``\n\n    ``pip install --upgrade pip``\n\n    ``poetry env use $(pyenv which python)``\n\n    ``pip install poetry``\n\n\n- Install the project\n\n    ``poetry install --no-root --extras "extras"``\n\n\nWhere extras may be in: `sqla`, `db`, `redis`, `rmq`, `sqlite`, `pg`, `mysql`, `scheduler`, `fastapi`, `test`\n\n\n- Start the docker environment for development\n\n    ``docker-compose up -d``\n\n\n- Run tests\n\n    ``poetry run pytest``\n\n\n- Stop the docker environment\n\n    ``docker-compose down``\n\n\n- Uninstall the project\n\n    ``pip uninstall dropland -y``\n',
    'author': 'Max Plutonium',
    'author_email': 'plutonium.max@gmail.com',
    'maintainer': 'Max Plutonium',
    'maintainer_email': 'plutonium.max@gmail.com',
    'url': 'https://gitlab.com/thegamma/dropland',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import contextlib
import enum
from datetime import timedelta
from typing import Any, Optional

from dropland.data.models.cache import MethodCache
from dropland.data.models.nosql import NoSqlModel, NoSqlProxyModel
from dropland.data.serializers import Serializer, Deserializer
from dropland.data.serializers.pickle import PickleSerializer, PickleDeserializer
from .engine import RedisEngine


class RedisCacheType(int, enum.Enum):
    SIMPLE = 0
    HASH = 1


class RedisModel(NoSqlModel):
    class Meta(NoSqlModel.Meta):
        _redis_engine = None

    def __init_subclass__(
            cls, redis_engine: RedisEngine = None,
            cache_type: RedisCacheType = RedisCacheType.SIMPLE,
            serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            ttl_enabled: bool = True,
            **kwargs):
        from .cache import SimpleRedisModelCache, HashRedisModelCache

        if not redis_engine:
            return super().__init_subclass__(**kwargs)

        serializer = serializer or PickleSerializer()
        deserializer = deserializer or PickleDeserializer()

        protocol_class = SimpleRedisModelCache \
            if cache_type == RedisCacheType.SIMPLE \
            else HashRedisModelCache

        cache_protocol = protocol_class(
            redis_engine, cls.__name__,
            serializer, deserializer,
            ttl_enabled=ttl_enabled
        )

        cls.Meta._redis_engine = redis_engine

        super().__init_subclass__(
            cache_protocol=cache_protocol,
            serializer=serializer,
            deserializer=deserializer,
            **kwargs
        )

    def get_id_value(self) -> Any:
        raise NotImplementedError

    # noinspection PyProtectedMember
    @classmethod
    def get_engine(cls) -> 'RedisEngine':
        return cls.Meta._redis_engine

    @classmethod
    @contextlib.asynccontextmanager
    async def _session_context(cls):
        async with cls.get_engine().session() as session:
            yield session


class RedisProxyModel(NoSqlProxyModel, RedisModel):
    class Meta(NoSqlProxyModel.Meta, RedisModel.Meta):
        pass

    def __init_subclass__(cls, redis_engine: RedisEngine = None, **kwargs):
        super().__init_subclass__(redis_engine=redis_engine, **kwargs)
        cls.Meta._redis_engine = redis_engine

    def get_id_value(self) -> Any:
        raise NotImplementedError


class RedisMethodCache(MethodCache):
    def __init__(
            self, redis_engine: RedisEngine, model_name: str,
            cache_type: RedisCacheType = RedisCacheType.SIMPLE,
            serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            ttl: Optional[timedelta] = None):
        from .cache import SimpleRedisModelCache, HashRedisModelCache

        serializer = serializer or PickleSerializer()
        deserializer = deserializer or PickleDeserializer()

        protocol_class = SimpleRedisModelCache \
            if cache_type == RedisCacheType.SIMPLE \
            else HashRedisModelCache

        cache_protocol = protocol_class(
            redis_engine, model_name,
            serializer, deserializer,
            ttl_enabled=True
        )

        super().__init__(cache_protocol, ttl)
        self._redis_engine = redis_engine
        self._cache_type = cache_type

    @contextlib.asynccontextmanager
    async def _session_context(self):
        async with self._redis_engine.session() as session:
            yield session
